from BeautifulSoup import BeautifulSoup, Comment
import sys, os, re

ERROR_MESSAGE = "<html><body><p>Could not parse input</p></body></html>"

# Derived from the HTML5 specification: 
#   http://www.w3schools.com/html5/html5_reference.asp
WHITELISTED_ELEMENTS = [
    'a', 'abbr', 'address', 'area', 'article', 'aside', 'b',
    'base', 'bdo', 'blockquote', 'body', 'br', 'canvas',
    'caption', 'center', 'cite', 'code', 'col', 'colgroup',
    'command', 'datalist', 'dd', 'del', 'dfn', 'div', 'dl',
    'dt', 'em', 'fieldset', 'figcaption', 'figure',
    'font', 'footer', 'form', 'h1', 'h2',
    'h3', 'h4', 'h5', 'h6', 'head', 'header', 'hr', 'html',
    'i', 'img', 'ins', 'keygen', 'kbd', 'label',
    'legend', 'li', 'map', 'mark', 'menu',
    'meter', 'nav', 'noscript', 'ol', 'optgroup', 'option',
    'output', 'p', 'param', 'pre', 'progress', 'q', 's', 'samp',
    'section', 'select', 'small', 'big', 'span', 'strike',
    'strong', 'sub', 'summary', 'sup', 'table', 'tbody',
    'td', 'textarea', 'tfoot', 'th', 'thead', 'time', 'title',
    'tr', 'ul', 'var'
    ]

# Couldn't find a comprehensive list of all available attributes; lifted the list instead from:
#   https://code.google.com/p/html5lib/source/browse/python/src/html5lib/sanitizer.py?r=79be9cc384d91a5d89fed3d777f274500d26e943
WHITELISTED_ATTRIBUTES = [
    'abbr', 'accept', 'accept-charset', 'accesskey',
    'action', 'align', 'alt', 'autocomplete', 'autofocus', 'axis',
    'balance', 'bgcolor', 'bgproperties', 'border',
    'bordercolor', 'bordercolordark', 'bordercolorlight', 'bottompadding',
    'cellpadding', 'cellspacing', 'ch', 'challenge', 'char', 'charoff',
    'choff', 'charset', 'checked', 'cite', 'class', 'clear', 'color',
    'cols', 'colspan', 'compact', 'contenteditable', 'controls', 'coords',
    'data', 'datafld', 'datapagesize', 'datasrc', 'datetime', 'default',
    'delay', 'dir', 'disabled', 'draggable', 'dynsrc', 'enctype', 'end',
    'face', 'for', 'form', 'frame', 'galleryimg', 'gutter', 'headers',
    'height', 'hidefocus', 'hidden', 'high', 'href', 'hreflang', 'hspace',
    'icon', 'id', 'inputmode', 'ismap', 'keytype', 'label', 'leftspacing',
    'lang', 'list', 'longdesc', 'loop', 'loopcount', 'loopend',
    'loopstart', 'low', 'lowsrc', 'max', 'maxlength', 'media', 'method',
    'min', 'multiple', 'name', 'nohref', 'noshade', 'nowrap', 'open',
    'optimum', 'pattern', 'point-size', 'prompt', 'pqg',
    'radiogroup', 'readonly', 'rel', 'repeat-max', 'repeat-min',
    'replace', 'required', 'rev', 'rightspacing', 'rows', 'rowspan',
    'rules', 'scope', 'selected', 'shape', 'size', 'span', 'src', 'start',
    'step', 'summary', 'suppress', 'tabindex', 'target',
    'template', 'title', 'toppadding', 'type', 'unselectable', 'usemap',
    'urn', 'valign', 'value', 'variable', 'volume', 'vspace', 'vrml',
    'width', 'wrap'
]
    

# This is largely speculative; unfamiliar with what other protocols are viable
WHITELISTED_PROTOCOLS = [
    'http', 'https', 'ftp', 'sftp','#'
    ]

# Derived from HTML5 specification of attributes that require URIs:
#   http://simon.html5.org/html-elements
REFERENCE_ATTRIBUTES = [
    'href', 'src', 'action', 'cite', 'base', 'background', 'dynsrc', 'lowsrc'
    ]

NUMERIC_ATTRIBUTES = [
    'bgcolor', 'color', 'colspan', 'height', 'width', 'size', 'bordercolor',
    'bordercolorlight', 'bordercolordark', 'bottompadding', 'cellpadding',
    'cellspacing'
    ]

class Sanitizer:

    # Attempt to parse the HTML provided; may cause parser to crash
    def parseHtml(self, html):
        try:
            parsed_html = BeautifulSoup(html)
            return parsed_html
        except:
            sys.exit(1)

    # Verify numeric value only
    def safeNumericAttribute(self, element, attribute, value):
        if attribute not in NUMERIC_ATTRIBUTES:
            return True
        elif re.search('[^A-Fa-f\d#]',value):
            return False
        else:
            return True

    # Verify attribute associated with a given element is whitelisted      
    def safeAttribute(self, element, attribute, value):
        if attribute in WHITELISTED_ATTRIBUTES:
            return True
        else:
            return False

    # Verify element in DOM tree is whitelisted
    def safeElement(self, element):
        if element.name not in WHITELISTED_ELEMENTS:
            return False
        else:
            return True

    # Verify a src, href, or similar reference points to a whitelisted protocol
    def safeReference(self, element, attribute, value):

        # Skip attributes which cant be URLs; automatically safe
        if attribute not in REFERENCE_ATTRIBUTES:
            return True
        else:
            is_valid_protocol = False
            for protocol in WHITELISTED_PROTOCOLS:
                if value.startswith(protocol):
                    is_valid_protocol = True
            return is_valid_protocol

    # Sanitizie an arbitrary HTML input
    def sanitizeHtml(self, html):
    
        parsed_html = self.parseHtml(html)

        # Walk the DOM tree to remove suspicious elements and attributes
        dropped_elements = []
        for element in parsed_html.findAll(True):

            if not self.safeElement(element):
                dropped_elements.append(element)
            else:
                for attribute, value in element._getAttrMap().items():
                    if not self.safeReference(element, attribute, value):
                        dropped_elements.append(element)
                    elif not self.safeAttribute(element, attribute, value):
                        del element[attribute]
                    elif not self.safeNumericAttribute(element, attribute, value):
                        del element[attribute]
  
        # Remove all comments; prevents browser-specific javascript/style hacks
        comments = parsed_html.findAll(text=lambda text:isinstance(text, Comment))
        for comment in comments:
            dropped_elements.append(comment)

        # Remove unsafe elements
        for element in dropped_elements:
            if element is not None:
                element.extract()

        # Finally return the html structure
        return parsed_html.prettify()

def verifyInput(html):
    if len(html) <= 0:
        sys.exit(1)

def main():
    try:
        html = sys.stdin.read()
        verifyInput(html)
        sanitizer = Sanitizer()
        
        safehtml = sanitizer.sanitizeHtml(html)
        print safehtml
    except:
        print ERROR_MESSAGE
        sys.exit(1)

if __name__ == "__main__":
    main()
